# -*- coding: utf-8 -*-
# wasp_backup/file_archiver.py
#
# Copyright (C) 2017 the wasp-backup authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-backup.
#
# wasp-backup is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-backup is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-backup.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_backup.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_backup.version import __status__

import io

from wasp_general.verify import verify_type, verify_value

from wasp_backup.cipher import WBackupCipher
from wasp_backup.core import WBackupMeta
from wasp_backup.archiver import WBasicArchiveCreator


class WFileArchiveCreator(WBasicArchiveCreator):

	@verify_type('paranoid', archive_path=str, io_write_rate=(float, int, None))
	@verify_value('paranoid', archive_path=lambda x: len(x) > 0, io_write_rate=lambda x: x is None or x > 0)
	@verify_type(cipher=(WBackupCipher, None), compression_mode=(WBackupMeta.Archive.CompressionMode, None))
	def __init__(
		self, backup_source, archive_path, logger, stop_event=None, io_write_rate=None, compression_mode=None,
		cipher=None, buffer_size=io.DEFAULT_BUFFER_SIZE
	):
		WBasicArchiveCreator.__init__(
			self, archive_path, logger, stop_event=stop_event, io_write_rate=io_write_rate,
			compression_mode=compression_mode, cipher=cipher
		)
		self.__backup_source = backup_source
		self.__buffer_size = buffer_size

	def backup_source(self):
		return self.__backup_source

	def buffer_size(self):
		return self.__buffer_size

	def write_archive(self, fo, archive):
		backup_source = self.backup_source()
		buffer_size = self.buffer_size()

		read_buffer = backup_source.read(buffer_size)
		while len(read_buffer) > 0:
			fo.write(read_buffer)
			read_buffer = backup_source.read(buffer_size)

	def meta(self):
		result = WBasicArchiveCreator.meta(self)
		result[WBackupMeta.Archive.MetaOptions.archived_files] = self.backup_source()
		return result



# -*- coding: utf-8 -*-
# wasp_backup/file_backup.py
#
# Copyright (C) 2017 the wasp-backup authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-backup.
#
# wasp-backup is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-backup is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-backup.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_backup.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_backup.version import __status__

from enum import Enum

from wasp_general.command.enhanced import WCommandArgumentDescriptor

from wasp_backup.cipher import WBackupCipher
from wasp_backup.inside_tar_archiver import WLVMArchiveCreator
from wasp_backup.command_common import __common_args__, WCreateBackupCommand


class WFileBackupCommand(WCreateBackupCommand):

	class SnapshotUsage(Enum):
		auto = 'auto'
		forced = 'forced'
		disabled = 'disabled'

	__command__ = 'file-backup'
	__description__ = 'create backup archive of files and directories'

	__arguments__ = (
		__common_args__['backup-archive'],
		__common_args__['input-files'],
		WCommandArgumentDescriptor(
			'sudo', flag_mode=True,
			help_info='use "sudo" command for privilege promotion. "sudo" may be used for snapshot '
			'creation, partition mounting and un-mounting'
		),
		WCommandArgumentDescriptor(
			'snapshot', help_info='whether to create snapshot before backup or not. '
			'One of: "auto" (backup will try to make snapshot for input files), '
			'"forced" (if snapshot can not be created - backup will fail), '
			'"disabled" (backup will not try to create a snapshot)',
			casting_helper=WCommandArgumentDescriptor.EnumArgumentHelper(SnapshotUsage),
			default_value=SnapshotUsage.auto.value
		),
		WCommandArgumentDescriptor(
			'snapshot-volume-size', meta_var='fraction_size',
			help_info='snapshot volume size as fraction of original volume size',
			casting_helper=WCommandArgumentDescriptor.FloatArgumentCastingHelper(
				validate_fn=lambda x: x > 0
			)
		),
		WCommandArgumentDescriptor(
			'snapshot-mount-dir', meta_var='mount_path',
			help_info='path where snapshot volume should be mount. It is random directory by default'
		),
		__common_args__['compression'],
		__common_args__['password'],
		__common_args__['cipher_algorithm'],
		__common_args__['io-write-rate'],
		__common_args__['copy-to'],
		__common_args__['copy-fail'],
		__common_args__['notify-app']
	)

	def _exec(self, command_arguments, **command_env):
		compression_mode = None
		if 'compression' in command_arguments.keys():
			compression_mode = command_arguments['compression']

		cipher = None
		if 'password' in command_arguments:
			cipher = WBackupCipher(
				command_arguments['cipher_algorithm'], command_arguments['password']
			)

		snapshot_size = None
		if 'snapshot-volume-size' in command_arguments.keys():
			snapshot_size = command_arguments['snapshot-volume-size']

		snapshot_mount_dir = None
		if 'snapshot-mount-dir' in command_arguments.keys():
			snapshot_mount_dir = command_arguments['snapshot-mount-dir']

		io_write_rate = None
		if 'io-write-rate' in command_arguments.keys():
			io_write_rate = command_arguments['io-write-rate']

		backup_archive = command_arguments['backup-archive']

		archiver = WLVMArchiveCreator(
			backup_archive, self.logger(), *command_arguments['input-files'],
			compression_mode=compression_mode, sudo=command_arguments['sudo'], cipher=cipher,
			io_write_rate=io_write_rate, stop_event=self.stop_event()
		)

		snapshot_disabled = (command_arguments['snapshot'] == WFileBackupCommand.SnapshotUsage.disabled)
		snapshot_force = (command_arguments['snapshot'] == WFileBackupCommand.SnapshotUsage.forced)

		self.set_archiver(archiver)
		return self._create_backup(
			command_arguments,
			disable_snapshot=snapshot_disabled,
			snapshot_force=snapshot_force,
			snapshot_size=snapshot_size,
			mount_directory=snapshot_mount_dir
		)



# -*- coding: utf-8 -*-
# wasp_backup/file_backup.py
#
# Copyright (C) 2017 the wasp-backup authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-backup.
#
# wasp-backup is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-backup is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-backup.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_backup.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_backup.version import __status__

from wasp_general.command.enhanced import WCommandArgumentDescriptor

from wasp_backup.cipher import WBackupCipher
from wasp_backup.popen_archiver import WPopenArchiveCreator
from wasp_backup.command_common import __common_args__, WCreateBackupCommand


class WProgramBackupCommand(WCreateBackupCommand):

	__command__ = 'program-backup'
	__description__ = 'create backup archive of stdout of program'

	__arguments__ = (
		__common_args__['backup-archive'],
		WCommandArgumentDescriptor(
			'input-program', required=True, multiple_values=False, meta_var='program_command',
			help_info='program which output will be backed up'
		),
		__common_args__['compression'],
		__common_args__['password'],
		__common_args__['cipher_algorithm'],
		__common_args__['io-write-rate'],
		__common_args__['copy-to'],
		__common_args__['copy-fail'],
		__common_args__['notify-app']
	)

	def _exec(self, command_arguments, **command_env):
		compression_mode = None
		if 'compression' in command_arguments.keys():
			compression_mode = command_arguments['compression']

		cipher = None
		if 'password' in command_arguments:
			cipher = WBackupCipher(
				command_arguments['cipher_algorithm'], command_arguments['password']
			)

		io_write_rate = None
		if 'io-write-rate' in command_arguments.keys():
			io_write_rate = command_arguments['io-write-rate']

		backup_archive = command_arguments['backup-archive']
		archiver = WPopenArchiveCreator(
			command_arguments['input-program'], backup_archive, self.logger(),
			compression_mode=compression_mode, cipher=cipher, io_write_rate=io_write_rate,
			stop_event=self.stop_event()
		)
		self.set_archiver(archiver)
		return self._create_backup(command_arguments)



# -*- coding: utf-8 -*-
# wasp_backup/retention.py
#
# Copyright (C) 2017 the wasp-backup authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-backup.
#
# wasp-backup is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-backup is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-backup.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_backup.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_backup.version import __status__

import re
import json
from datetime import datetime, timedelta
from enum import Enum
from pytz import timezone

from wasp_general.command.enhanced import WCommandArgumentDescriptor, WCommandArgumentRelationship
from wasp_general.command.result import WPlainCommandResult
from wasp_general.network.clients.base import WCommonNetworkClientCapability
from wasp_general.network.clients.collection import __default_client_collection__
from wasp_general.uri import WURI
from wasp_general.datetime import local_tz

from wasp_backup.command_common import WBackupCommand, __common_args__
from wasp_backup.core import WBackupMeta
from wasp_backup.notify import notify


class WRetentionBackupCommand(WBackupCommand):

	class AgeHelper(Enum):
		name_parsing = 'name-parsing'
		archive_meta = 'archive-meta'
		modification_time = 'modification-time'
		creation_time = 'creation-time'

	class PeriodKeepFilter:

		def __init__(self, from_dt, period_value, period_modifier, archive_number):
			self.__from_dt = from_dt
			self.__reduce_fn = None

			if period_modifier in ['M', 'H', 'd', 'w']:
				period = int(period_value) * 60  # at least a minute
				if period_modifier != 'M':
					period = period * 60  # at least an hour
					if period_modifier != 'H':
						period = period * 24  # at least a day
						if period_modifier == 'w':
							period = period * 7
				self.__reduce_fn = lambda: self.__from_dt - timedelta(seconds=period)
			elif period_modifier == 'm':

				def reduce_fn():
					year = self.__from_dt.year()
					if self.__from_dt.month == 1:
						year -= 1
						month = 12
					else:
						month = self.__from_dt.month - 1

					return datetime(
						year=year, month=month, day=self.__from_dt.day,
						hour=self.__from_dt.hour, minute=self.__from_dt.minute,
						second=self.__from_dt.second, microsecond=self.__from_dt.microsecond
					)
				self.__reduce_fn = reduce_fn
			elif period_modifier == 'y':
				self.__reduce_fn = lambda: datetime(
					year=self.__from_dt.year - 1, month=self.__from_dt.month,
					day=self.__from_dt.day, hour=self.__from_dt.hour,
					minute=self.__from_dt.minute, second=self.__from_dt.second,
					microsecond=self.__from_dt.microsecond
				)
			else:
				raise ValueError('Unknown period_modifier was specified')

			self.__to_dt = self.__reduce_fn()
			self.__archive_number = int(archive_number)

		def __call__(self, check_item):
			if self.__archive_number <= 0:
				return False

			archive_name, creation_date = check_item

			if creation_date > self.__from_dt:
				return False
			if creation_date < self.__to_dt:
				return False

			self.__archive_number -= 1
			self.__from_dt = self.__to_dt
			self.__to_dt = self.__reduce_fn()
			return True

	__command__ = 'retention'
	__description__ = 'rotate archive backups that resides locally or on a remote location'

	__arguments__ = [
		WCommandArgumentDescriptor(
			'backup-location', required=True, multiple_values=False, meta_var='location',
			help_info='Location (network or directory) that has multiple backups to rotate'
		),

		WCommandArgumentDescriptor(
			'period-keep', required=True, multiple_values=True, meta_var='period@number_of_copies',
			help_info='parameter speicifies set of archives that should be kept from deleting. During '
			' the specified value only ONE archive will be kept. The value should be input in the '
			'following format: [period value][period modifier]@[number of periods]. Where '
			'[period modifier] is one of "M", "H", "d", "w", "m", "y" that represent "minute", "hour", '
			'"day", "week", "month", "year" respectively. [number of periods] is number of periods. '
			'For example "2d10" means that there will be 10 archives one archive per 2 days that will '
			'be kept. May be specified multiple times. In that case it works as a union of a single '
			'"period-keep" result',
			casting_helper=WCommandArgumentDescriptor.RegExpArgumentHelper('(\d+)([MHdwmy])@(\d+)')
		),

		WCommandArgumentDescriptor(
			'timezone', required=True, multiple_values=False, meta_var='timezone_name',
			help_info='timezone that will be apply to age-helper. If archive-meta-age-helper is specified '
			'then "UTC" should be used. For other helpers "local" is a good choice'
		),

		WCommandArgumentDescriptor(
			'minimum-archives', required=True, multiple_values=False, meta_var='archives_count',
			help_info='Number of archives that will be kept even if they are expired (the youngest '
			'archives will be selected).',
			casting_helper=WCommandArgumentDescriptor.IntegerArgumentCastingHelper(
				validate_fn=lambda x: x > 0
			)
		),

		WCommandArgumentDescriptor(
			'archive-selection', required=False, multiple_values=False, meta_var='regexp',
			help_info='regular expression that is used to select archives from the backup location',
			default_value='^.*$'
		),

		WCommandArgumentDescriptor(
			'name-parser-age-helper', flag_mode=True, help_info='defines method that will be used for '
			'archive to determine its age. This one will define age by parsing an archive name (dumb but '
			'fast)'
		),

		WCommandArgumentDescriptor(
			'archive-meta-age-helper', flag_mode=True, help_info='defines method that will be used for '
			'archive to determine its age. This one will define age from meta data from an archive (smart '
			'but slow, because archive may be downloaded and because meta data needs to be extracted)'
		),

		WCommandArgumentDescriptor(
			'modification-time-age-helper', flag_mode=True, help_info='defines method that will be used for '
			'archive to determine its age. This one will define age from archive modification time (not '
			'all of the location types supports this type of information, so it may be unavailable)'
		),

		WCommandArgumentDescriptor(
			'creation-time-age-helper', flag_mode=True,
			help_info='defines method that will be used for '
			'archive to determine its age. This one will define age from archive creation time (not '
			'all of the location types supports this type of information, so it may be unavailable)'
		),

		WCommandArgumentDescriptor(
			'date-format', required=False, multiple_values=False, meta_var='format',
			help_info='defines format of a date that will be used to find archive age (format has '
			'the same syntax as strptime python function)'
		),

		WCommandArgumentDescriptor(
			'download-location', required=False, multiple_values=False, meta_var='directory_path',
			help_info='directory where archives should be downloaded to in order to fecth archive meta '
			'data (used with "archive-meta-age-helper" flag)', default_value='/var/tmp'
		),
		__common_args__['notify-app']
	]

	__relationships__ = [
		WCommandArgumentRelationship(
			WCommandArgumentRelationship.Relationship.one_of,
			'name-parser-age-helper',
			'archive-meta-age-helper',
			'modification-time-age-helper',
			'creation-time-age-helper'
		),
		WCommandArgumentRelationship(
			WCommandArgumentRelationship.Relationship.requirement,
			'name-parser-age-helper',
			'date-format'
		),
	]

	def _exec(self, command_arguments, **command_env):
		location = command_arguments['backup-location']
		uri = WURI.parse(location)
		network_client = __default_client_collection__.open(uri)
		archives = network_client.request(WCommonNetworkClientCapability.list_dir)

		archive_selection_re = re.compile(command_arguments['archive-selection'])
		re_selected_archives = tuple(filter(lambda x: archive_selection_re.match(x) is not None, archives))

		if command_arguments['timezone'] != 'local':
			tz = timezone(command_arguments['timezone'])
		else:
			tz = local_tz()
		now = datetime.now(tz=tz)

		age_helper = self.__age_helper(command_arguments, network_client, tz)
		archive_ages = [(x, age_helper(x)) for x in re_selected_archives]

		archive_ages = list(filter(lambda x: (now - x[1]).total_seconds() > 0, archive_ages))  # remove list fn
		archive_ages.sort(key=lambda x: (now - x[1]).total_seconds())
		sorted_archives = [x[0] for x in archive_ages]

		archive_to_keep = set()
		for period_keep in command_arguments['period-keep']:
			archive_to_keep.update(filter(
				WRetentionBackupCommand.PeriodKeepFilter(now, *period_keep), archive_ages
			))

		keep_archives = [x[0] for x in archive_to_keep]

		extra_archives_required = command_arguments['minimum-archives'] - len(keep_archives)
		if extra_archives_required > 0:
			for i in range(len(sorted_archives)):
				archive_name = sorted_archives[i]
				if archive_name not in keep_archives:
					keep_archives.append(archive_name)
					extra_archives_required -= 1

					if extra_archives_required <= 0:
						break

		files_to_remove = set(sorted_archives).difference(keep_archives)

		for file_name in files_to_remove:
			network_client.request(WCommonNetworkClientCapability.remove_file, file_name)

		if 'notify-app' in command_arguments:
			notify(
				{
					WBackupMeta.RetentionNotificationOptions.retention_location: location,
					WBackupMeta.RetentionNotificationOptions.kept_archives: list(keep_archives),
					WBackupMeta.RetentionNotificationOptions.removed_archives: list(files_to_remove)
				},
				command_arguments['notify-app'],
				encode_strict_cls=(WBackupMeta.RetentionNotificationOptions)
			)

		return WPlainCommandResult(
			'Archives deleted - %i, archives kept - %i' %
			(len(files_to_remove), len(set(re_selected_archives).difference(files_to_remove)))
		)

	def __age_helper(self, command_arguments, network_client, tz):
		if command_arguments['name-parser-age-helper'] is True:
			return self.__name_parser_helper(tz, command_arguments['date-format'])
		elif command_arguments['archive-meta-age-helper'] is True:
			return self.__archive_meta_helper(tz, network_client, command_arguments['download-location'])
		elif command_arguments['modification-time-age-helper'] is True:
			return self.__modification_time_helper(tz, network_client)
		elif command_arguments['creation-time-age-helper'] is True:
			return self.__creation_time_helper(tz, network_client)
		raise ValueError('Unknown helper name is specified')

	def __name_parser_helper(self, tz, date_format):
		def helper(archive):
			return datetime.strptime(archive, date_format).replace(tzinfo=tz)
		return helper

	def __archive_meta_helper(self, tz, network_client, download_location):
		# TODO: implement this
		raise NotImplementedError('Not ready')

	def __modification_time_helper(self, tz, network_client):
		# TODO: implement this
		raise NotImplementedError('Not ready')

	def __creation_time_helper(self, tz, network_client):
		# TODO: implement this
		raise NotImplementedError('Not ready')



# -*- coding: utf-8 -*-
# wasp_backup/version.py
#
# Copyright (C) 2017 the wasp-backup authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-backup.
#
# wasp-backup is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-backup is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-backup.  If not, see <http://www.gnu.org/licenses/>.

import os
import json
import subprocess


with open(os.path.join(os.path.dirname(__file__), 'package.json'), 'r') as f:
	__package_data__ = json.load(f)


def package_version():

	result = __package_data__['numeric_version']
	if __package_data__['dev_suffix'] is True:
		try:

			cwd = os.getcwd()
			try:
				os.chdir(os.path.dirname(__file__))
				# check_output is function that defined in python3 and python2
				with open(os.devnull, 'w') as f:  # python2 does not have subprocess.DEVNULL
					output = subprocess.check_output(["git", "rev-parse", "HEAD"], stderr=f)

				# check_output in python3 returns bytes, python2 - str
				if isinstance(output, bytes) is True:
					output = output.decode()
				result += '.dev%s' % output[:7]
			finally:
				os.chdir(cwd)

		except (subprocess.CalledProcessError, OSError):
			result += '--'
	return result


__author__ = __package_data__['author']
__email__ = __package_data__['author_email']
__credits__ = __package_data__['credits']
__license__ = __package_data__['license']
__copyright__ = __package_data__['copyright']
__status__ = __package_data__['status']
__version__ = package_version()



# -*- coding: utf-8 -*-
# wasp_backup/inside_tar_archiver.py
#
# Copyright (C) 2017 the wasp-backup authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-backup.
#
# wasp-backup is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-backup is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-backup.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_backup.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_backup.version import __status__

import os
import uuid
import tempfile

from wasp_general.verify import verify_type, verify_value
from wasp_general.os.linux.lvm import WLogicalVolume
from wasp_general.os.linux.mounts import WMountPoint

from wasp_backup.cipher import WBackupCipher
from wasp_backup.core import WBackupMeta
from wasp_backup.archiver import WBasicInsideTarArchiveCreator


class WInsideTarArchiveCreator(WBasicInsideTarArchiveCreator):

	@verify_type('paranoid', archive_path=str, cipher=(WBackupCipher, None))
	@verify_type('paranoid', compression_mode=(WBackupMeta.Archive.CompressionMode, None))
	@verify_type('paranoid', io_write_rate=(float, int, None))
	@verify_type(backup_sources=str, abs_path=bool)
	@verify_value('paranoid', archive_path=lambda x: len(x) > 0, io_write_rate=lambda x: x is None or x > 0)
	@verify_value(backup_sources=lambda x: len(x) > 0)
	def __init__(
		self, archive_path, logger, *backup_sources, compression_mode=None, cipher=None, stop_event=None,
		io_write_rate=None, abs_path=False
	):
		WBasicInsideTarArchiveCreator.__init__(
			self, archive_path, logger, compression_mode=compression_mode, cipher=cipher, stop_event=stop_event,
			io_write_rate=io_write_rate
		)

		self.__backup_sources = list(backup_sources)
		self.__abs_path = abs_path
		self.__last_file = None

	def backup_sources(self):
		return self.__backup_sources.copy()

	def abs_path(self):
		return self.__abs_path

	def last_file(self):
		return self.__last_file

	def inside_filename(self):
		result = WBackupMeta.Archive.__basic_inside_file_name__ + '.tar'
		compression_mode = self.compression_mode()
		if compression_mode is not None:
			result += '.' + compression_mode.value
		return result

	def archive(self):
		self.__last_file = None
		WBasicInsideTarArchiveCreator.archive(self)

	def _populate_archive(self, tar_archive):
		def last_file_tracking(tarinfo):
			self.__last_file = tarinfo.name
			return tarinfo

		for entry in self.backup_sources():
			if self.abs_path() is True:
				entry = os.path.abspath(entry)
			tar_archive.add(entry, recursive=True, filter=last_file_tracking)

	def meta(self):
		result = WBasicInsideTarArchiveCreator.meta(self)
		result.update({
			WBackupMeta.Archive.MetaOptions.inside_tar: True,
			WBackupMeta.Archive.MetaOptions.archived_files: self.backup_sources(),
		})
		return result


class WLVMArchiveCreator(WInsideTarArchiveCreator):

	@verify_type('paranoid', archive_path=str, backup_sources=str)
	@verify_type('paranoid', compression_mode=(WBackupMeta.Archive.CompressionMode, None))
	@verify_type('paranoid', cipher=(WBackupCipher, None), io_write_rate=(float, int, None))
	@verify_value('paranoid', archive_path=lambda x: len(x) > 0, backup_sources=lambda x: len(x) > 0)
	@verify_value('paranoid', io_write_rate=lambda x: x is None or x > 0)
	@verify_type(sudo=bool)
	def __init__(
		self, archive_path, logger, *backup_sources, compression_mode=None, sudo=False, cipher=None, stop_event=None,
		io_write_rate=None
	):
		WInsideTarArchiveCreator.__init__(
			self, archive_path, logger, *backup_sources, compression_mode=compression_mode, cipher=cipher,
			stop_event=stop_event, io_write_rate=io_write_rate, abs_path=True
		)
		self.__sudo = sudo
		self.__logical_volume_uuid = None
		self.__snapshot = False

	def sudo(self):
		return self.__sudo

	@verify_type(disable_snapshot=bool, snapshot_force=bool, snapshot_size=(int, float, None), mount_directory=(str, None))
	@verify_type(mount_fs=(str, None), mount_options=(list, tuple, set, None))
	def archive(
		self, disable_snapshot=False, snapshot_force=False, snapshot_size=None, mount_directory=None,
		mount_fs=None, mount_options=None
	):
		if disable_snapshot is True and snapshot_force is True:
			raise ValueError('Conflict flags "disable_snapshot" and "snapshot_force" was specified')

		self.__logical_volume_uuid = None
		self.__snapshot = False

		logical_volume = None
		backup_sources = self.backup_sources()

		if len(backup_sources) == 0:
			if snapshot_force is True:
				raise RuntimeError('Unable to create snapshot for empty archive')
			else:
				WBasicInsideTarArchiveCreator.archive(self)
				return

		if disable_snapshot is False:
			for source in backup_sources:
				lv = WLogicalVolume.logical_volume(source, sudo=self.sudo())
				if lv is None:
					if snapshot_force is True:
						raise RuntimeError('Unable to create snapshot for non-LVM volume')
					logical_volume = None
					break
				if logical_volume is None:
					logical_volume = lv
				elif os.path.realpath(logical_volume.volume_path()) == os.path.realpath(lv.volume_path()):
					pass
				else:
					if snapshot_force is True:
						raise RuntimeError(
							'Unable to create snapshot - files reside on different volumes'
						)
					logical_volume = None
					break

		if logical_volume is None:
			if snapshot_force is True:
				raise RuntimeError('Unable to create snapshot for unknown reason')
			WBasicInsideTarArchiveCreator.archive(self)
			return

		if snapshot_size is None:
			snapshot_size = WBackupMeta.LVMSnapshot.__default_snapshot_size__

		snapshot_suffix = '-snapshot-%s' % str(uuid.uuid4())
		snapshot_volume = None
		remove_directory = False
		directory_mounted = False
		current_cwd = os.getcwd()

		try:
			snapshot_volume = logical_volume.create_snapshot(snapshot_size, snapshot_suffix)
			self.__logical_volume_uuid = logical_volume.uuid()
			self.__snapshot = True

			if mount_directory is None:
				mount_directory = tempfile.mkdtemp(
					suffix=snapshot_suffix,
					prefix=WBackupMeta.LVMSnapshot.__mount_directory_prefix__
				)
				remove_directory = True
			if mount_options is None:
				mount_options = []
			mount_options.insert(0, 'ro')

			WMountPoint.mount(
				snapshot_volume.volume_path(), mount_directory, fs=mount_fs, options=mount_options,
				sudo=self.sudo()
			)
			directory_mounted = True

			mount_directory_length = len(mount_directory)
			for i in range(len(backup_sources)):
				backup_sources[i] = backup_sources[i][mount_directory_length:]

			os.chdir(mount_directory)
			WInsideTarArchiveCreator.archive(self)

		except Exception:
			self.__logical_volume_uuid = None
			self.__snapshot = False
			raise
		finally:
			os.chdir(current_cwd)

			if directory_mounted is True:
				WMountPoint.umount(snapshot_volume.volume_path(), sudo=self.sudo())

			if remove_directory is True:
				os.removedirs(mount_directory)

			if snapshot_volume is not None:
				snapshot_volume.remove_volume()

	def meta(self):
		meta = WInsideTarArchiveCreator.meta(self)
		meta[WBackupMeta.Archive.MetaOptions.snapshot_used] = self.__snapshot
		meta[WBackupMeta.Archive.MetaOptions.original_lv_uuid] = \
			self.__logical_volume_uuid if self.__logical_volume_uuid is not None else ''
		return meta



# -*- coding: utf-8 -*-
# wasp_backup/check.py
#
# Copyright (C) 2017 the wasp-backup authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-backup.
#
# wasp-backup is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-backup is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-backup.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_backup.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_backup.version import __status__

from wasp_general.command.enhanced import WCommandArgumentDescriptor
from wasp_general.command.result import WPlainCommandResult


from wasp_backup.command_common import WBackupCommand
from wasp_backup.archiver import WArchiveIntegrityChecker


class WCheckBackupCommand(WBackupCommand):

	__command__ = 'check'
	__description__ = 'check backup archive for integrity'

	__arguments__ = [
		WCommandArgumentDescriptor(
			'backup-archive', required=True, multiple_values=False, meta_var='archive_path',
			help_info='backup file to check'
		),
		WCommandArgumentDescriptor(
			'io-read-rate', meta_var='maximum reading rate',
			help_info='use this parameter to limit disk I/O load (bytes per second). You can use '
			'suffixes like "K" for kibibytes, "M" for mebibytes, "G" for gibibytes, "T" for tebibytes for '
			'convenience ', casting_helper=WCommandArgumentDescriptor.DataSizeArgumentHelper()
		)
	]

	def checker(self):
		return self.__checker

	def _exec(self, command_arguments, **command_env):
		archive = command_arguments['backup-archive']
		io_read_rate = None
		if 'io-read-rate' in command_arguments.keys():
			io_read_rate = command_arguments['io-read-rate']

		try:
			self.__checker = WArchiveIntegrityChecker(
				archive, self.logger(), stop_event=self.stop_event(), io_read_rate=io_read_rate
			)
			result, original_hash, calculated_hash = self.__checker.check_archive()
		finally:
			self.__checker = None

		if result is True:
			return WPlainCommandResult('Archive "%s" is OK' % archive)
		return WPlainCommandResult.error(
			'Archive "%s" is corrupted. Calculated hash - "%s". Original hash - "%s"' %
			(archive, calculated_hash, original_hash)
		)



# -*- coding: utf-8 -*-
# wasp_backup/notify.py
#
# Copyright (C) 2018 the wasp-backup authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-backup.
#
# wasp-backup is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-backup is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-backup.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_backup.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_backup.version import __status__

import os
import shlex
import sys
import tempfile
from wasp_backup.core import WBackupMeta, WBackupMetaProvider


def notify(notify_data, notify_app, encode_strict_cls=None):
	first_fork_pid = os.fork()
	if first_fork_pid == 0:
		second_fork_pid = os.fork()
		if second_fork_pid == 0:
			binary_meta = WBackupMetaProvider.encode_meta(
				notify_data, strict_cls=encode_strict_cls
			)

			meta_tempfile = tempfile.NamedTemporaryFile(delete=False)
			meta_tempfile.write(binary_meta)
			meta_tempfile.close()

			app_tokens = shlex.split(notify_app)

			os.execvpe(
				app_tokens[0],
				app_tokens,
				{WBackupMeta.__notification_env_var_name__: meta_tempfile.name}
			)
		else:
			sys.exit(0)
	else:
		os.waitpid(first_fork_pid, 0)



# -*- coding: utf-8 -*-
# wasp_backup/popen_archiver.py
#
# Copyright (C) 2017 the wasp-backup authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-backup.
#
# wasp-backup is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-backup is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-backup.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_backup.archiver import WBasicArchiveCreator
# noinspection PyUnresolvedReferences
from wasp_backup.version import __status__

import shlex
import subprocess

from wasp_backup.file_archiver import WFileArchiveCreator
from wasp_backup.core import WBackupMeta


class WPopenArchiveCreator(WFileArchiveCreator):

	def write_archive(self, fo, archive):
		with subprocess.Popen(shlex.split(self.backup_source()), stdout=subprocess.PIPE) as pipe:
			buffer_size = self.buffer_size()
			read_buffer = pipe.stdout.read(buffer_size)
			while len(read_buffer) > 0:
				fo.write(read_buffer)
				read_buffer = pipe.stdout.read(buffer_size)

	def meta(self):
		result = WFileArchiveCreator.meta(self)
		if WBackupMeta.Archive.MetaOptions.archived_files in result:
			result.pop(WBackupMeta.Archive.MetaOptions.archived_files)
		result[WBackupMeta.Archive.MetaOptions.archived_program] = self.backup_source()
		return result



# -*- coding: utf-8 -*-
# wasp_backup/archiver.py
#
# Copyright (C) 2017 the wasp-backup authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-backup.
#
# wasp-backup is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-backup is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-backup.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_backup.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_backup.version import __status__

import os
import tarfile
import json
import math
from time import mktime
from datetime import datetime

from wasp_general.verify import verify_type, verify_value
from wasp_general.io import WWriterChainLink, WReaderChainLink, WThrottlingReader, WResponsiveWriter, WResponsiveIO
from wasp_general.io import WResponsiveReader, WHashCalculationReader, WDiscardReaderResult, WReaderChain, WGzipReader
from wasp_general.io import WBzip2Reader

from wasp_backup.cipher import WBackupCipher
from wasp_backup.core import WBackupMeta
from wasp_backup.io import WMetaTarPatcher, WArchiverThrottlingWriter, WArchiverHashCalculationWriter
from wasp_backup.io import WArchiverAESCipher, WArchiverThrottlingReader
from wasp_backup.io import WArchiverWriterChain, WExtractorReaderChain, WBackupMetaProvider, WBasicArchiverIO
from wasp_backup.io import WArchiverDataCounter


"""

archiving:

(lvm snapshot)
	|
	|
	| ->  tar( + compression) -> (encryption ->) hashing
							|
							|-> single tar archive -> (throttling ->) file object
							|                               |
archive meta information -------------------------------|     (may be automatic split because of target fs limitation?)
											|
											|-> splitter object
												|
												| -> file object 1
												|
												| -> file object 2
												|
												...
												| -> file object n
"""


class WBasicArchiveCreator(WBasicArchiverIO, WBackupMetaProvider):

	@verify_type('paranoid', archive_path=str, io_write_rate=(float, int, None))
	@verify_value('paranoid', archive_path=lambda x: len(x) > 0, io_write_rate=lambda x: x is None or x > 0)
	@verify_type(cipher=(WBackupCipher, None), compression_mode=(WBackupMeta.Archive.CompressionMode, None))
	def __init__(
		self, archive_path, logger, stop_event=None, io_write_rate=None, compression_mode=None,
		cipher=None
	):
		WBasicArchiverIO.__init__(self, archive_path, logger, stop_event=stop_event, io_rate=io_write_rate)
		WBackupMetaProvider.__init__(self)
		self.__compression_mode = compression_mode
		self.__cipher = cipher
		self.__writer_chain = None
		self.__last_archive_creation_time = None

	def io_write_rate(self):
		return self.io_rate()

	def compression_mode(self):
		return self.__compression_mode

	def cipher(self):
		return self.__cipher

	def file_object(self):
		return self.__writer_chain

	def archiving_details(self):
		if self.__writer_chain is not None:
			return self.__writer_chain.status()

	def inside_filename(self):
		result = WBackupMeta.Archive.__basic_inside_file_name__
		compression_mode = self.compression_mode()
		if compression_mode is not None:
			result += '.' + compression_mode.value
		return result

	def write_chain(self):
		inside_archive_name = self.inside_filename()

		chain = [
			open(self.archive_path(), mode='wb'),
			WWriterChainLink(WArchiverThrottlingWriter, write_limit=self.io_write_rate()),
			WWriterChainLink(
				WMetaTarPatcher, inside_archive_name, self, compression_mode=self.compression_mode()
			),
			WWriterChainLink(WArchiverDataCounter),
			WWriterChainLink(WArchiverHashCalculationWriter)
		]

		cipher = self.cipher()
		if cipher is not None:
			chain.append(WWriterChainLink(WArchiverAESCipher, cipher))

		stop_event = self.stop_event()
		if stop_event is not None:
			chain.append(WWriterChainLink(WResponsiveWriter, stop_event))

		return WArchiverWriterChain(*chain)

	def archive(self):
		archive_path = self.archive_path()
		self.__writer_chain = self.write_chain()
		self.__last_archive_creation_time = self.__utc_unix_time()
		archive_instance = self.__writer_chain.instance(WMetaTarPatcher)

		try:
			self.write_archive(self.__writer_chain, archive_instance)
			self.__writer_chain.flush()
			archive_instance.patch()
			self.__writer_chain.close()
			self.logger().info('Archive "%s" was created and patched successfully' % archive_path)

		except WResponsiveIO.IOTerminated:
			os.unlink(archive_path)
			self.logger().error(
				'Unable to create archive "%s" - task terminated, changes discarded' % archive_path
			)
			return
		except Exception:
			os.unlink(archive_path)
			self.logger().error('Unable to create archive "%s". Changes discarded' % archive_path)
			raise

	@classmethod
	def __utc_unix_time(cls):
		utc_datetime = datetime.utcnow()
		utc_unix_time = mktime(utc_datetime.timetuple())
		return math.floor(utc_unix_time)

	def write_archive(self, fo, archive):
		pass

	def meta(self):
		result = self.__writer_chain.meta() if self.__writer_chain is not None else {}

		compression_mode = self.compression_mode()
		if compression_mode is not None:
			compression_mode = compression_mode.value

		result.update({
			WBackupMeta.Archive.MetaOptions.inside_filename: self.inside_filename(),
			WBackupMeta.Archive.MetaOptions.compression_mode: compression_mode,
			WBackupMeta.Archive.MetaOptions.creation_time: self.__last_archive_creation_time
		})
		return result


class WBasicInsideTarArchiveCreator(WBasicArchiveCreator):

	@verify_type('paranoid', archive_path=str, io_write_rate=(float, int, None))
	@verify_type('paranoid', cipher=(WBackupCipher, None), compression_mode=(WBackupMeta.Archive.CompressionMode, None))
	@verify_value('paranoid', archive_path=lambda x: len(x) > 0, io_write_rate=lambda x: x is None or x > 0)
	def __init__(self, archive_path, logger, compression_mode=None, cipher=None, stop_event=None, io_write_rate=None):
		WBasicArchiveCreator.__init__(
			self, archive_path, logger, stop_event=stop_event, io_write_rate=io_write_rate,
			compression_mode=compression_mode, cipher=cipher
		)

		self.__compression_mode = compression_mode
		self.__cipher = cipher
		self.__writer_chain = None

	def write_archive(self, fo, archive):
		tar = tarfile.open(fileobj=fo, mode='w:')
		self._populate_archive(tar)
		fo.flush()

		data_written = archive.data_written()
		padding_size = archive.record_size(data_written + (tarfile.BLOCKSIZE * 2)) - data_written
		fo.write(archive.padding(padding_size))

	def _populate_archive(self, tar_archive):
		pass


class WBasicArchiveExtractor(WBasicArchiverIO):

	@verify_type('paranoid', archive_path=str, io_read_rate=(float, int, None))
	@verify_value('paranoid', archive_path=lambda x: len(x) > 0, io_read_rate=lambda x: x is None or x > 0)
	def __init__(self, archive_path, logger, stop_event=None, io_read_rate=None):
		WBasicArchiverIO.__init__(self, archive_path, logger, stop_event=stop_event, io_rate=io_read_rate)

	def io_read_rate(self):
		return self.io_rate()

	@verify_type(file_name=str)
	@verify_value(file_name=lambda x: len(x) > 0)
	def open_file(self, file_name):
		chain = [
			open(self.archive_path(), 'rb'),
			WReaderChainLink(WThrottlingReader, throttling_to=self.io_read_rate()),
		]

		stop_event = self.stop_event()
		if stop_event is not None:
			chain.append(WReaderChainLink(WResponsiveReader, stop_event))

		reader_chain = WReaderChain(*chain)

		tar = tarfile.open(fileobj=reader_chain, mode='r:')
		extracted_file = tar.extractfile(file_name)
		if extracted_file is None:
			raise RuntimeError(
				'File "%s" was not found in archive "%s"' % (file_name, self.archive_path())
			)
		return extracted_file

	def open_meta(self):
		return self.open_file(WBackupMeta.Archive.__meta_filename__)


class WArchiveIntegrityChecker(WBasicArchiveExtractor):

	@verify_type('paranoid', archive_path=str, io_read_rate=(float, int, None))
	@verify_value('paranoid', archive_path=lambda x: len(x) > 0, io_read_rate=lambda x: x is None or x > 0)
	def __init__(self, archive_path, logger, stop_event=None, io_read_rate=None):
		WBasicArchiveExtractor.__init__(self, archive_path, logger, stop_event=stop_event, io_read_rate=io_read_rate)
		self.__reader_chain = None

	def reader_chain(self):
		return self.__reader_chain

	def check_details(self):
		if self.__reader_chain is not None:
			return self.__reader_chain.status()

	def check_archive(self):
		try:
			meta_file_data = self.open_meta()
			json_raw_data = meta_file_data.read()
			meta_file_data.close()
			json_data = json.loads(json_raw_data.decode())
			inside_archive_name = json_data[WBackupMeta.Archive.MetaOptions.inside_filename.value]

			chain = [self.open_file(inside_archive_name)]

			if WBackupMeta.Archive.MetaOptions.compression_mode.value in json_data:
				compression_mode = json_data[WBackupMeta.Archive.MetaOptions.compression_mode.value]
				if compression_mode is not None:
					if compression_mode == WBackupMeta.Archive.CompressionMode.gzip.value:
						chain.append(WReaderChainLink(WGzipReader))
					elif compression_mode == WBackupMeta.Archive.CompressionMode.bzip2.value:
						chain.append(WReaderChainLink(WBzip2Reader))
					else:
						raise RuntimeError(
							'Unsupported compression mode spotted: "%s"' % compression_mode
						)

			chain.extend([
				WReaderChainLink(
					WHashCalculationReader,
					json_data[WBackupMeta.Archive.MetaOptions.hash_algorithm.value]
				),
				WReaderChainLink(WArchiverThrottlingReader),
				WReaderChainLink(WDiscardReaderResult)
			])
			self.__reader_chain = WExtractorReaderChain(*chain)
			self.__reader_chain.read()
			calc_instance = self.__reader_chain.instance(WHashCalculationReader)
			self.__reader_chain.close()

			original_hash = json_data[WBackupMeta.Archive.MetaOptions.hash_value.value].upper()
			calculated_hash = calc_instance.hexdigest().upper()
			return original_hash == calculated_hash, original_hash, calculated_hash
		except WResponsiveIO.IOTerminated:
			self.logger().error(
				'Unable to check archive "%s" - task terminated' % self.archive_path()
			)
			return
		finally:
			self.__reader_chain = None


"""
__openssl_mode_re__ = re.compile('aes-([0-9]+)-(.+)')
bits, mode = __openssl_mode_re__.search(cipher.lower()).groups()
key_size = int(int(bits) / 8)
mode = 'AES-%s' % mode.upper()
'''

'''
import hmac
import hashlib
import Crypto.Protocol.KDF
fn = lambda x,y: hmac.new(x,msg=y,digestmod=hashlib.sha256).digest()
salt = b'\x01\x02\x03\x04\x05\x06\x07\x08'
Crypto.Protocol.KDF.PBKDF2('password', salt, prf=fn)

echo -en password | nettle-pbkdf2 -i 1000 -l 16 --hex-salt 0102030405060708
openssl enc -aes-256-cbc -d -in 1.tar.gz.aes -out 1.tar.gz -K \
	"c057f2deac4cba660f5463b8346ee67961948a598e0f4f72e7ad46d2ffeecd39" -iv "4084a32c07fb808e8dfc679c3cde6480" \
	-nosalt -nopad
"""



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# wasp-backup.py
#
# Copyright (C) 2016 the wasp-backup authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-backup.
#
# Wasp-backup is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-backup is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-backup.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_backup.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_backup.version import __status__

import sys
import os
from logging import getLogger

from wasp_general.command.command import WCommandSet, WCommandProto, WCommandPrioritizedSelector, WCommand
from wasp_general.command.result import WPlainCommandResult

from wasp_backup.file_backup import WFileBackupCommand
from wasp_backup.check import WCheckBackupCommand
from wasp_backup.program_backup import WProgramBackupCommand
from wasp_backup.retention import WRetentionBackupCommand


class WCommandHelp(WCommand):

	__help_info__ = '''This utility is able to create file or program backup, to check archive integrity and, \
and is able to rotate archives that resides locally or on a remote location.
Syntax: %s <main_command> [<command argument 1> <command argument 2> <command argument 3>...]

''' % sys.argv[0]

	def __init__(self, command_selector):
		WCommand.__init__(self, 'help')
		self.__command_selector = command_selector

	def _exec(self, *command_tokens, **command_env):
		result = self.__help_info__
		for command in self.__command_selector.commands():
			if isinstance(command, WCommandHelp) is True:
				continue
			result += 'Command "%s" is able to: %s\n' % (command.__command__, command.__description__)
			result += command.command_help()
			result += '\n\n'

		return WPlainCommandResult(result)


if __name__ == '__main__':

	logger = getLogger(os.path.basename(sys.argv[0]))

	command_set = WCommandSet(command_selector=WCommandPrioritizedSelector())
	command_set.commands().add_prioritized(WCommandHelp(command_set), 30)
	command_set.commands().add_prioritized(WFileBackupCommand(logger), 50)
	command_set.commands().add_prioritized(WProgramBackupCommand(logger), 50)
	command_set.commands().add_prioritized(WCheckBackupCommand(logger), 50)
	command_set.commands().add_prioritized(WRetentionBackupCommand(logger), 50)
	print(command_set.exec(WCommandProto.join_tokens(*(sys.argv[1:]))))



# -*- coding: utf-8 -*-
# wasp_backup/apps.py
#
# Copyright (C) 2017 the wasp-backup authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-backup.
#
# wasp-backup is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-backup is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-backup.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_backup.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_backup.version import __status__

from wasp_general.task.scheduler.task_source import WInstantTaskSource
from wasp_general.cli.formatter import na_formatter

from wasp_launcher.core import WAppsGlobals
from wasp_launcher.core_scheduler import WSchedulerTaskSourceInstaller, WLauncherTaskSource
from wasp_launcher.core_broker import WCommandKit, WBrokerCommand, WResponsiveBrokerCommand

from wasp_backup.core import WBackupMeta
from wasp_backup.file_backup import WFileBackupCommand
from wasp_backup.check import WCheckBackupCommand
from wasp_backup.program_backup import WProgramBackupCommand
from wasp_backup.retention import WRetentionBackupCommand


class WResponsiveCreateBackupCommand(WResponsiveBrokerCommand):

	class FileBackupCommand(WFileBackupCommand, WBrokerCommand):

		def __init__(self):
			WFileBackupCommand.__init__(self, WAppsGlobals.log)
			WBrokerCommand.__init__(
				self, self.command_token(), *self.argument_descriptors(),
				relationships=self.relationships()
			)

		def brief_description(self):
			return self.__description__

	class ScheduledTask(WResponsiveBrokerCommand.ScheduledTask):

		def state_details(self):
			archiver = self.basic_command().archiver()
			if archiver is not None:
				return 'Archiving is not running. May be finalizing'

			result = 'Archiving file: %s' % na_formatter(archiver.last_file())
			details = archiver.archiving_details()
			if details is not None:
				result += '\n' + details
			return result

		def thread_started(self):
			self.basic_command().stop_event(self.stop_event())
			WResponsiveBrokerCommand.ScheduledTask.thread_started(self)

	__task_source_name__ = WBackupMeta.__task_source_name__
	__scheduler_instance__ = WBackupMeta.__scheduler_instance_name__

	def __init__(self):
		WResponsiveBrokerCommand.__init__(
			self, WResponsiveCreateBackupCommand.FileBackupCommand(),
			scheduled_task_cls=WResponsiveCreateBackupCommand.ScheduledTask
		)


class WResponsiveCheckBackupCommand(WResponsiveBrokerCommand):

	class CheckBackupCommand(WCheckBackupCommand, WBrokerCommand):

		def __init__(self):
			WCheckBackupCommand.__init__(self, WAppsGlobals.log)
			WBrokerCommand.__init__(
				self, self.command_token(), *self.argument_descriptors(),
				relationships=self.relationships()
			)

		def brief_description(self):
			return self.__description__

	class ScheduledTask(WResponsiveBrokerCommand.ScheduledTask):

		def state_details(self):
			checker = self.basic_command().checker()
			if checker is not None:
				details = checker.check_details()
				if details is not None:
					return '\n' + details

			return 'Checking is not running. May be finalizing'

		def thread_started(self):
			self.basic_command().stop_event(self.stop_event())
			WResponsiveBrokerCommand.ScheduledTask.thread_started(self)

	__task_source_name__ = WBackupMeta.__task_source_name__
	__scheduler_instance__ = WBackupMeta.__scheduler_instance_name__

	def __init__(self):
		WResponsiveBrokerCommand.__init__(
			self, WResponsiveCheckBackupCommand.CheckBackupCommand(),
			scheduled_task_cls=WResponsiveCheckBackupCommand.ScheduledTask
		)


class WResponsiveProgramBackupCommand(WResponsiveBrokerCommand):

	class ProgramBackupCommand(WProgramBackupCommand, WBrokerCommand):

		def __init__(self):
			WProgramBackupCommand.__init__(self, WAppsGlobals.log)
			WBrokerCommand.__init__(
				self, self.command_token(), *self.argument_descriptors(),
				relationships=self.relationships()
			)

		def brief_description(self):
			return self.__description__

	class ScheduledTask(WResponsiveBrokerCommand.ScheduledTask):

		def state_details(self):
			archiver = self.basic_command().archiver()
			if archiver is not None:
				return 'Archiving is not running. May be finalizing'

			result = ''
			details = archiver.archiving_details()
			if details is not None:
				result += '\n' + details
			return result

		def thread_started(self):
			self.basic_command().stop_event(self.stop_event())
			WResponsiveBrokerCommand.ScheduledTask.thread_started(self)

	__task_source_name__ = WBackupMeta.__task_source_name__
	__scheduler_instance__ = WBackupMeta.__scheduler_instance_name__

	def __init__(self):
		WResponsiveBrokerCommand.__init__(
			self, WResponsiveProgramBackupCommand.ProgramBackupCommand(),
			scheduled_task_cls=WResponsiveProgramBackupCommand.ScheduledTask
		)


class WResponsiveRetentionCommand(WResponsiveBrokerCommand):

	class RetentionCommand(WRetentionBackupCommand, WBrokerCommand):

		def __init__(self):
			WRetentionBackupCommand.__init__(self, WAppsGlobals.log)
			WBrokerCommand.__init__(
				self, self.command_token(), *self.argument_descriptors(),
				relationships=self.relationships()
			)

		def brief_description(self):
			return self.__description__

	class ScheduledTask(WResponsiveBrokerCommand.ScheduledTask):

		def state_details(self):
			# TODO: make it more descriptive!
			return 'Task is running?!'

		def thread_started(self):
			# TODO: make it more responsive!
			#self.basic_command().stop_event(self.stop_event())
			WResponsiveBrokerCommand.ScheduledTask.thread_started(self)

	__task_source_name__ = WBackupMeta.__task_source_name__
	__scheduler_instance__ = WBackupMeta.__scheduler_instance_name__

	def __init__(self):
		WResponsiveBrokerCommand.__init__(
			self, WResponsiveRetentionCommand.RetentionCommand(),
			scheduled_task_cls=WResponsiveRetentionCommand.ScheduledTask
		)


class WBackupBrokerCommandKit(WCommandKit):

	__registry_tag__ = 'com.binblob.wasp-backup.broker-commands'

	@classmethod
	def description(cls):
		return 'backup creation/restoring commands'

	@classmethod
	def commands(cls):
		return (
			WResponsiveCreateBackupCommand(),
			WResponsiveCheckBackupCommand(),
			WResponsiveProgramBackupCommand(),
			WResponsiveRetentionCommand()
		)


class WBackupSchedulerInstaller(WSchedulerTaskSourceInstaller):

	__scheduler_instance__ = WBackupMeta.__scheduler_instance_name__

	class InstantTaskSource(WInstantTaskSource, WLauncherTaskSource):

		__task_source_name__ = WBackupMeta.__task_source_name__

		def __init__(self, scheduler):
			WInstantTaskSource.__init__(self, scheduler)
			WLauncherTaskSource.__init__(self)

		def name(self):
			return self.__task_source_name__

		def description(self):
			return 'Backup tasks from broker'


	__registry_tag__ = 'com.binblob.wasp-backup.scheduler.sources'

	def sources(self):
		return WBackupSchedulerInstaller.InstantTaskSource,



# -*- coding: utf-8 -*-
# wasp_backup/cipher.py
#
# Copyright (C) 2017 the wasp-backup authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-backup.
#
# wasp-backup is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-backup is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-backup.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_backup.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_backup.version import __status__

from wasp_general.crypto.aes import WAES, WAESMode, WZeroPadding
from wasp_general.crypto.kdf import WPBKDF2
from wasp_general.crypto.hmac import WHMAC

from wasp_backup.core import WBackupMeta


class WBackupCipher:

	__pbkdf2_iterations_count__ = 10000
	__hmac_hash_generator_name__ = 'SHA256'

	def __init__(self, cipher_name, password):
		self.__cipher_name = cipher_name
		aes_key_size, aes_mode = WAESMode.parse_cipher_name(cipher_name)
		init_seq_length = WAESMode.init_sequence_length(aes_key_size, aes_mode)
		kdf = WPBKDF2(
			password, derived_key_length=init_seq_length,
			hmac=WHMAC(self.__hmac_hash_generator_name__), iterations_count=self.__pbkdf2_iterations_count__
		)
		self.__salt = kdf.salt()
		self.__aes = WAES(WAESMode(aes_key_size, aes_mode, kdf.derived_key(), padding=WZeroPadding()))

	def cipher_name(self):
		return self.__cipher_name

	def salt(self):
		return self.__salt

	def aes_cipher(self):
		return self.__aes

	def meta(self):
		salt = ''
		for salt_byte in self.salt():
			salt += "{:02x}".format(salt_byte)

		return {
			WBackupMeta.Archive.MetaOptions.cipher_algorithm:
				self.cipher_name(),
			WBackupMeta.Archive.MetaOptions.pbkdf2_salt:
				salt,
			WBackupMeta.Archive.MetaOptions.pbkdf2_iterations_count:
				self.__pbkdf2_iterations_count__,
			WBackupMeta.Archive.MetaOptions.pbkdf2_prf:
				('HMAC-%s' % self.__hmac_hash_generator_name__)
		}



# -*- coding: utf-8 -*-
# wasp_backup/command_common.py
#
# Copyright (C) 2017 the wasp-backup authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-backup.
#
# wasp-backup is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-backup is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-backup.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_backup.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_backup.version import __status__

import os
import sys
import shlex
from datetime import datetime
import tempfile

from wasp_general.verify import verify_type
from wasp_general.uri import WURI
from wasp_general.network.clients.base import WCommonNetworkClientCapability
from wasp_general.network.clients.collection import __default_client_collection__
from wasp_general.command.enhanced import WCommandArgumentDescriptor
from wasp_general.crypto.aes import WAESMode
from wasp_general.command.result import WPlainCommandResult
from wasp_general.command.enhanced import WEnhancedCommand

from wasp_backup.core import WBackupMeta, WBackupMetaProvider
from wasp_backup.notify import notify


class WCompressionArgumentHelper(WCommandArgumentDescriptor.ArgumentCastingHelper):

	def __init__(self):
		WCommandArgumentDescriptor.ArgumentCastingHelper.__init__(
			self, casting_fn=self.cast_string
		)

	@staticmethod
	@verify_type(value=str)
	def cast_string(value):
		value = value.lower()
		if value == 'gzip':
			return WBackupMeta.Archive.CompressionMode.gzip
		elif value == 'bzip2':
			return WBackupMeta.Archive.CompressionMode.bzip2
		elif value == 'disabled':
			return
		else:
			raise ValueError('Invalid compression value')


def cipher_name_validation(cipher_name):
	try:
		if WAESMode.parse_cipher_name(cipher_name) is not None:
			return True
	except ValueError:
		pass
	return False


__common_args__ = {
	'backup-archive': WCommandArgumentDescriptor(
		'backup-archive', required=True, multiple_values=False, meta_var='archive_path',
		help_info='backup file path'
	),

	'input-files': WCommandArgumentDescriptor(
		'input-files', required=True, multiple_values=True, meta_var='input_path',
		help_info='files or directories to backup'
	),

	'input-program': WCommandArgumentDescriptor(
		'input-program', required=True, multiple_values=False, meta_var='program_command',
		help_info='program which output will be backed up'
	),

	'compression': WCommandArgumentDescriptor(
		'compression', meta_var='compression_type',
		help_info='compression option. One of: "gzip", "bzip2" or "disabled". It is disabled by default',
		casting_helper=WCompressionArgumentHelper()
	),

	'password': WCommandArgumentDescriptor(
		'password', meta_var='encryption_password',
		help_info='password to encrypt backup. Backup is not encrypted by default'
	),

	'cipher_algorithm': WCommandArgumentDescriptor(
		'cipher_algorithm', meta_var='algorithm_name',
		help_info='cipher that will be used for encrypt (backup will not be encrypted if password was not '
		'set). It is "AES-256-CBC" by default',
		casting_helper=WCommandArgumentDescriptor.StringArgumentCastingHelper(
			validate_fn=cipher_name_validation
		),
		default_value='AES-256-CBC'
	),

	'io-write-rate': WCommandArgumentDescriptor(
		'io-write-rate', meta_var='maximum writing rate',
		help_info='use this parameter to limit disk I/O load (bytes per second). You can use '
		'suffixes like "K" for kibibytes, "M" for mebibytes, "G" for gibibytes, "T" for tebibytes for '
		'convenience ', casting_helper=WCommandArgumentDescriptor.DataSizeArgumentHelper()
	),

	'copy-to': WCommandArgumentDescriptor(
		'copy-to', meta_var='URL', help_info='Location to copy backup archive to'
	),

	'copy-fail': WCommandArgumentDescriptor(
		'copy-fail', flag_mode=True, help_info='If specified, then backup will fail if copy operation fails. '
		'(But local archive would not be deleted any way)',
	),

	'notify-app': WCommandArgumentDescriptor(
		'notify-app', meta_var='app_path', help_info='Application that will be called as a handler'
	),
}


# noinspection PyAbstractClass
class WBackupCommand(WEnhancedCommand):

	__command__ = None

	__arguments__ = tuple()
	__relationships__ = None

	def __init__(self, logger):
		WEnhancedCommand.__init__(
			self, self.__command__, *self.__arguments__, relationships=self.__relationships__
		)
		self.__logger = logger
		self.__stop_event = None

	def logger(self):
		return self.__logger

	def stop_event(self, value=None):
		if value is not None:
			self.__stop_event = value
		return self.__stop_event


# noinspection PyAbstractClass
class WCreateBackupCommand(WBackupCommand):

	class UploadFailed(Exception):
		pass

	def __init__(self, logger):
		WBackupCommand.__init__(self, logger)
		self.__archiver = None

	def archiver(self):
		return self.__archiver

	def set_archiver(self, value):
		self.__archiver = value

	def _create_backup(self, command_arguments, *args, **kwargs):
		archiver = self.archiver()
		if archiver is None:
			raise RuntimeError('Archiver must be set before call')

		try:
			backup_started_at = datetime.utcnow()
			archiver.archive(*args, **kwargs)
			backup_duration = (datetime.utcnow() - backup_started_at).seconds

			copy_to = None
			if 'copy-to' in command_arguments.keys():
				copy_to = command_arguments['copy-to']

			notify_app = None
			if 'notify-app' in command_arguments.keys():
				notify_app = command_arguments['notify-app']

			copy_fail = False
			if 'copy-fail' in command_arguments.keys():
				copy_fail = command_arguments['copy-fail']

			if copy_to is None:
				return self.__handle_backup_result(
					'Archive "%s" was created successfully' % archiver.archive_path(),
					notify_app=notify_app,
					backup_duration=backup_duration
				)

			copy_result, copy_duration = self.__copy(archiver.archive_path(), copy_to)

			if copy_result is True:
				backup_result = \
					'Archive "%s" was created and uploaded successfully' % archiver.archive_path()
			elif copy_fail is False:
				backup_result = \
					'Archive "%s" was created successfully. But it fails to upload archive to ' \
					'destination' % archiver.archive_path()
			else:
				raise WCreateBackupCommand.UploadFailed(
					'Unable to upload archive "%s"' % archiver.archive_path()
				)

			return self.__handle_backup_result(
				backup_result,
				notify_app=notify_app,
				backup_duration=backup_duration,
				copy_to=copy_to,
				copy_complete=copy_result,
				copy_duration=copy_duration
			)

		finally:
			self.set_archiver(None)

	def __copy(self, archive_path, copy_to):
		copy_started_at = datetime.utcnow()
		uri = WURI.parse(copy_to)
		dir_name, file_name = os.path.split(uri.path())
		uri.component(WURI.Component.path, dir_name)

		network_client = __default_client_collection__.open(uri)
		with open(archive_path, 'rb') as f:
			copy_result = network_client.request(WCommonNetworkClientCapability.upload_file, file_name, f)
		copy_duration = (datetime.utcnow() - copy_started_at).total_seconds()
		return copy_result, copy_duration

	def __handle_backup_result(
		self, str_result, notify_app=None, backup_duration=None, copy_to=None, copy_complete=None,
		copy_duration=None
	):
		archiver = self.archiver()
		if archiver is None:
			raise RuntimeError('Archiver must be set before call')

		if notify_app is None:
			return WPlainCommandResult(str_result)

		meta_data = archiver.meta()
		meta_data[WBackupMeta.BackupNotificationOptions.created_archive] = archiver.archive_path()
		meta_data[WBackupMeta.BackupNotificationOptions.backup_duration] = backup_duration
		meta_data[WBackupMeta.BackupNotificationOptions.total_archive_size] = os.stat(archiver.archive_path()).st_size
		meta_data[WBackupMeta.BackupNotificationOptions.copy_to] = copy_to
		meta_data[WBackupMeta.BackupNotificationOptions.copy_completion] = copy_complete
		meta_data[WBackupMeta.BackupNotificationOptions.copy_duration] = copy_duration

		notify(
			meta_data, notify_app,
			encode_strict_cls=(WBackupMeta.Archive.MetaOptions, WBackupMeta.BackupNotificationOptions)
		)
		return WPlainCommandResult(str_result)



# -*- coding: utf-8 -*-
# wasp_backup/io.py
#
# Copyright (C) 2017 the wasp-backup authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-backup.
#
# wasp-backup is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-backup is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-backup.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_backup.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_backup.version import __status__

import tarfile
import math
import io
import os
import gzip
import bz2
import time
import pwd
import grp
from datetime import datetime

from abc import ABCMeta, abstractmethod

from wasp_general.verify import verify_type, verify_value
from wasp_general.io import WAESWriter, WHashCalculationWriter, WWriterChain, WThrottlingWriter, WWriterChainLink
from wasp_general.io import WReaderChain, WThrottlingReader, WReaderChainLink, WDiscardWriterResult
from wasp_general.cli.formatter import data_size_formatter

from wasp_backup.core import WBackupMeta, WBackupMetaProvider, WArchiverIOStatusProvider


class WTarPatcher(io.BufferedWriter):

	__default_tar_mode__ = int('440', base=8)

	def __init__(
		self, archive, inside_file_name, patch_header=True, patch_tail=False, compression_mode=None
	):
		self.__original_archive = \
			open(archive, mode='wb', buffering=0) if isinstance(archive, str) is True else archive

		io.BufferedWriter.__init__(self, WDiscardWriterResult(self.__original_archive))

		self.__start_position = self.__original_archive.tell()
		self.__final_position = None
		self.__compression_writer = None

		if patch_header is True:
			self.__original_archive.write(self.tar_header(inside_file_name))

		self.__compression_mode = compression_mode
		if self.__compression_mode is not None:
			if self.__compression_mode == WBackupMeta.Archive.CompressionMode.gzip:
				archive = gzip.GzipFile(fileobj=self.__original_archive)
				self.__compression_writer = archive

			elif self.__compression_mode == WBackupMeta.Archive.CompressionMode.bzip2:
				archive = bz2.BZ2File(self.__original_archive, mode='wb')
				self.__compression_writer = archive
			else:
				raise RuntimeError('Invalid compression mode spotted')

		self.__inside_file_name = inside_file_name
		self.__patch_header = patch_header
		self.__patch_tail = patch_tail

		self.__data_written = 0

	def original_archive(self):
		return self.__original_archive

	def start_position(self):
		return self.__start_position

	def final_position(self):
		return self.__final_position

	def compression_mode(self):
		return self.__compression_mode

	def inside_file_size(self):
		final_position = self.final_position()
		if final_position is None:
			self.flush()
			if self.__compression_writer is not None:
				self.__compression_writer.flush()
			original_archive = self.original_archive()
			original_archive.flush()
			original_archive.seek(0, os.SEEK_END)
			final_position = self.__original_archive.tell()

		result = final_position - self.start_position()
		if self.patch_header() is True:
			result -= tarfile.BLOCKSIZE
		return result

	def inside_file_name(self):
		return self.__inside_file_name

	def patch_header(self):
		return self.__patch_header

	def patch_tail(self):
		return self.__patch_tail

	def data_written(self):
		return self.__data_written

	def write(self, b):
		self.__data_written += len(b)
		writer = self.__compression_writer if self.__compression_writer is not None else self.__original_archive
		writer.write(b)
		return len(b)

	def close(self):
		if self.__compression_writer is not None:
			self.__compression_writer.close()
		self.__original_archive.close()
		io.BufferedWriter.close(self)

	def patch(self):
		if self.__compression_writer is not None:
			self.__compression_writer.flush()
			self.__compression_writer.close()

		self.__original_archive.flush()
		self.__original_archive.seek(0, os.SEEK_END)
		self.__final_position = self.__original_archive.tell()

		if self.patch_tail():
			self._apply_tail_patch()

		if self.patch_header() is True:
			self._apply_header_patch()

	def alignment_padding(self):
		return self.block_size(self.data_written()) - self.data_written()

	def _apply_tail_patch(self):
		file_size = self.inside_file_size() + (tarfile.BLOCKSIZE * 2)
		if self.patch_header() is True:
			file_size += tarfile.BLOCKSIZE

		delta = self.record_size(file_size) - file_size
		self.__original_archive.write(self.padding(delta))

	def _apply_header_patch(self):
		tar_header = self.tar_header(self.inside_file_name(), size=self.inside_file_size())
		self.__original_archive.seek(self.start_position(), os.SEEK_SET)
		self.__original_archive.write(tar_header)

	@classmethod
	def tar_info(cls, name, size=None):
		tar_info = tarfile.TarInfo(name=name)
		if size is not None:
			tar_info.size = size
		tar_info.mtime = time.mktime(datetime.now().timetuple())
		tar_info.mode = cls.__default_tar_mode__
		tar_info.type = tarfile.REGTYPE
		tar_info.uid = os.getuid()
		tar_info.gid = os.getgid()
		tar_info.uname = pwd.getpwuid(tar_info.uid).pw_name
		tar_info.gname = grp.getgrgid(tar_info.gid).gr_name
		return tar_info

	@classmethod
	def tar_header(cls, name, size=None):
		return cls.tar_info(name, size=size).tobuf()

	@classmethod
	def align_size(cls, size, chunk_size):
		result = divmod(size, chunk_size)
		return (result[0] if result[1] == 0 else (result[0] + 1)) * chunk_size

	@classmethod
	def record_size(cls, size):
		return cls.align_size(size, tarfile.RECORDSIZE)

	@classmethod
	def block_size(cls, size):
		return cls.align_size(size, tarfile.BLOCKSIZE)

	@classmethod
	def padding(cls, padding_size):
		return tarfile.NUL * padding_size if padding_size > 0 else b''


class WMetaTarPatcher(WTarPatcher):

	def __init__(self, archive_path, inside_archive_name, meta_provider, compression_mode=None):
		WTarPatcher.__init__(
			self, archive_path, inside_archive_name, patch_tail=True, compression_mode=compression_mode
		)
		self.__meta_provider = meta_provider

	def meta_provider(self):
		return self.__meta_provider

	def _apply_tail_patch(self):
		original_archive = self.original_archive()
		inside_file_size = original_archive.tell() - self.start_position()
		inside_data_block_delta = self.block_size(inside_file_size) - inside_file_size
		original_archive.write(self.padding(inside_data_block_delta))
		inside_file_size += inside_data_block_delta

		meta_provider = self.meta_provider()
		meta_data = meta_provider.encode_meta(meta_provider.meta(), strict_cls=WBackupMeta.Archive.MetaOptions)

		if len(meta_data) > WBackupMeta.Archive.__maximum_meta_file_size__:
			raise RuntimeError('Meta data corrupted - too big')

		meta_header = self.tar_header(WBackupMeta.Archive.__meta_filename__, size=len(meta_data))
		original_archive.write(meta_header)
		inside_file_size += len(meta_header)

		original_archive.write(meta_data)
		inside_file_size += len(meta_data)

		meta_padding = self.block_size(len(meta_data)) - len(meta_data)
		inside_file_size += meta_padding
		original_archive.write(self.padding(meta_padding))

		delta = self.record_size(inside_file_size + (tarfile.BLOCKSIZE * 2)) - inside_file_size
		original_archive.write(self.padding(delta))

	@classmethod
	def process_meta(cls, meta):
		result = {}
		for meta_key, meta_value in meta.items():
			if isinstance(meta_key, WBackupMeta.Archive.MetaOptions) is False:
				raise TypeError('Invalid meta key spotted')
			result[meta_key.value] = meta_value
		return result


class WArchiverHashCalculationWriter(WHashCalculationWriter, WBackupMetaProvider):

	def __init__(self, raw):
		WHashCalculationWriter.__init__(self, raw, WBackupMeta.Archive.__hash_generator_name__)
		WBackupMetaProvider.__init__(self)

	def meta(self):
		return {
			WBackupMeta.Archive.MetaOptions.hash_algorithm:
				WBackupMeta.Archive.__hash_generator_name__,
			WBackupMeta.Archive.MetaOptions.hash_value:
				self.hexdigest()
		}


class WArchiverAESCipher(WAESWriter, WBackupMetaProvider):

	def __init__(self, raw, cipher):
		WAESWriter.__init__(self, raw, cipher.aes_cipher())
		WBackupMetaProvider.__init__(self)
		self.__meta = cipher.meta()

	def meta(self):
		return self.__meta


class WArchiverThrottlingWriter(WThrottlingWriter, WBackupMetaProvider, WArchiverIOStatusProvider):

	def __init__(self, raw, write_limit=None):
		WThrottlingWriter.__init__(self, raw, throttling_to=write_limit)
		WBackupMetaProvider.__init__(self)
		WArchiverIOStatusProvider.__init__(self)

	def meta(self):
		return {
			WBackupMeta.Archive.MetaOptions.io_write_rate: math.ceil(self.rate())
		}

	def status(self):
		result = 'Write rate: %s/sec\n' % data_size_formatter(math.ceil(self.rate()))
		result += 'Bytes processed: %i' % self.bytes_processed()
		return result


class WArchiverDataCounter(WThrottlingWriter, WBackupMetaProvider):

	def __init__(self, raw):
		WThrottlingWriter.__init__(self, raw)
		WBackupMetaProvider.__init__(self)

	def meta(self):
		return {
			WBackupMeta.Archive.MetaOptions.uncompressed_archive_size: self.bytes_processed()
		}


class WArchiverThrottlingReader(WThrottlingReader, WArchiverIOStatusProvider):

	def __init__(self, raw, read_limit=None):
		WThrottlingReader.__init__(self, raw, throttling_to=read_limit)
		WArchiverIOStatusProvider.__init__(self)

	def status(self):
		result = 'Read rate: %s/sec\n' % data_size_formatter(math.ceil(self.rate()))
		result += 'Bytes processed: %i' % self.bytes_processed()
		return result


class WArchiverStatus(metaclass=ABCMeta):

	def meta(self):
		result = {}
		for link in self:
			if isinstance(link, WBackupMetaProvider) is True:
				result.update(link.meta())
		return result

	def status(self):
		result = []
		for link in self:
			if isinstance(link, WArchiverIOStatusProvider) is True:
				result.append(link.status())
		if len(result) > 0:
			return '\n'.join(result)

	@abstractmethod
	def __iter__(self):
		raise NotImplementedError('This method is abstract')


class WArchiverWriterChain(WWriterChain, WArchiverStatus):

	@verify_type('paranoid', links=WWriterChainLink)
	def __init__(self, last_io_obj, *links):
		WWriterChain.__init__(self, last_io_obj, *links)
		WArchiverStatus.__init__(self)


class WExtractorReaderChain(WReaderChain, WArchiverStatus):

	@verify_type('paranoid', links=WReaderChainLink)
	def __init__(self, last_io_obj, *links):
		WReaderChain.__init__(self, last_io_obj, *links)
		WArchiverStatus.__init__(self)


class WBasicArchiverIO:

	@verify_type(archive_path=str, io_rate=(float, int, None))
	@verify_value(archive_path=lambda x: len(x) > 0, io_rate=lambda x: x is None or x > 0)
	def __init__(self, archive_path, logger, stop_event=None, io_rate=None):
		self.__archive_path = archive_path
		self.__logger = logger
		self.__stop_event = stop_event
		self.__io_rate = io_rate

	def archive_path(self):
		return self.__archive_path

	def logger(self):
		return self.__logger

	def io_rate(self):
		return self.__io_rate

	def stop_event(self, value=None):
		if value is not None:
			self.__stop_event = value
		return self.__stop_event



# -*- coding: utf-8 -*-
# wasp_backup/core.py
#
# Copyright (C) 2017 the wasp-backup authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-backup.
#
# wasp-backup is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-backup is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-backup.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_backup.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_backup.version import __status__

import json
from enum import Enum

from wasp_general.network.upload import WUploaderCollection, WFTPUploader


class WBackupMeta:

	class Archive:

		class CompressionMode(Enum):
			gzip = 'gz'
			bzip2 = 'bz2'

		class MetaOptions(Enum):
			creation_time = 'creation_time'  # unix time of archive creation (for UTC timezone)
			inside_filename = 'inside_filename'
			inside_tar = 'inside_tar'
			archived_files = 'archived_files'
			archived_program = 'archived_program'
			uncompressed_archive_size = 'uncompressed_archive_size'  # size of uncompressed data
			# (for inside_tar archive, this is a size of uncompressed inside tar, which is rounded to 10240)
			compression_mode = 'compression_mode'
			hash_algorithm = 'hash_algorithm'
			hash_value = 'hash_value'  # hash value of uncompressed inside archive (for
			# inside_tar archive, this is a hash of uncompressed inside tar)
			snapshot_used = 'snapshot_used'
			original_lv_uuid = 'original_lv_uuid'
			io_write_rate = 'io_write_rate'
			pbkdf2_salt = 'pbkdf2_salt'
			pbkdf2_prf = 'pbkdf2_prf'
			pbkdf2_iterations_count = 'pbkdf2_iterations_count'
			cipher_algorithm = 'cipher_algorithm'

		__meta_filename__ = 'meta.json'
		__maximum_meta_file_size__ = 50 * 1024 * 1024
		__basic_inside_file_name__ = 'archive'
		__file_mode__ = int('660', base=8)
		__hash_generator_name__ = 'MD5'

	class BackupNotificationOptions(Enum):
		created_archive = 'created_archive'
		backup_duration = 'backup_duration'
		copy_to = 'copy_to'
		copy_completion = 'copy_completion'
		copy_duration = 'copy_duration'
		total_archive_size = 'total_archive_size'

	class RetentionNotificationOptions(Enum):
		retention_location = 'retention_location'
		kept_archives = 'kept_archives'
		removed_archives = 'removed_archives'

	class LVMSnapshot:
		__default_snapshot_size__ = 0.1
		__mount_directory_prefix__ = 'wasp-backup-'

	__scheduler_instance_name__ = 'com.binblob.wasp-backup'
	__task_source_name__ = 'com.binblob.wasp-backup.scheduler.sources.instant_source'
	__notification_env_var_name__ = 'WASP_NOTIFICATION_META_FILE'

	__uploader_collection__ = WUploaderCollection(WFTPUploader())


class WBackupMetaProvider:

	def meta(self):
		return {}

	@classmethod
	def encode_meta(cls, meta, strict_cls=None):
		result = {}
		for meta_key, meta_value in meta.items():
			if strict_cls is not None:
				if isinstance(meta_key, strict_cls) is True:
					result[meta_key.value] = meta_value
				else:
					raise TypeError('Invalid meta key spotted')
			else:
				result[meta_key] = meta_value

		return json.dumps(result).encode()


class WArchiverIOStatusProvider:

	def status(self):
		return None
